/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.battle;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import org.xmloperator.reactive.Event;
import org.xmloperator.reactive.LaunchingContext;
import org.xmloperator.reactive.ReactiveCommunity;

import org.xmloperator.reactive.agent.AgentImpl;

/**
 * This Agent plays in the battle with cards.
 */
public final class BattlePlayerAgent extends AgentImpl {
	private final int playerNumber;
	private final BattleGameEvents gameEvents;
	private final long startInstant;
	private final List<Event> cardsInHand = new ArrayList<Event>();
	private final Random random = new Random();
	private final List<Event> playedCards = new ArrayList<Event>();
	
	public BattlePlayerAgent(int playerNumber, BattleGameEvents gameEvents, long startInstant) {
		if (playerNumber < 0 || startInstant < 1)
			throw new IllegalArgumentException();
		this.playerNumber = playerNumber;
		this.gameEvents = gameEvents;
		this.startInstant = startInstant;
		this.initializeCardsInHand();
		this.addParameter(playerNumber);
	}

	/**
	 * At the beginning of the game the player has n cards in his hands. They
	 * are marked "0" to "n-1". The "0" card is the strongest one. The "n-1"
	 * card is the lowest one. The cards are shuffled.
	 */
	private void initializeCardsInHand() {
		this.cardsInHand.clear();
		List<Event> cardEvents = new ArrayList<Event>();
		int n = this.gameEvents.cardCountByPlayer;
		for (int card = 0; card < n; card++)
			cardEvents.add(this.gameEvents
					.getCardEvent(this.playerNumber, card));
		this.winCards(cardEvents);
	}

	/**
	 * Wins the given played cards. The obtained cards are shuffled.
	 * @param cardEvents played cards.
	 */
	private void winCards(List<Event> cardEvents) {
		int n = cardEvents.size();
		for (int k = 0; k < n; k++) {
			int p = this.random.nextInt(n - k);
			this.cardsInHand.add(cardEvents.remove(p));
		}
	}
	
	/**
	 * Playing a card (always the first in the hand) consists in removing the
	 * card from his hand and comparing it with the ones of the other players.
	 * The greatest card get the whole cards. Equality needs the concerned
	 * players to play a new card.
	 * 
	 * @return true for continuing, false for terminating.
	 */
	private boolean playACard() {
		this.playedCards.clear();
		int playedZeroCardCount = 0;
		boolean isFirstCard = true;
		while (true) {
			if (this.cardsInHand.isEmpty()) {
				this.trace.println(this.entity.getCommunity().getInstant()
						+ " : " + this.entity + " has no more card.");
				return false;
			}
			/*
			 * Get the first card in the hand.
			 */
			Event myCardEvent = this.cardsInHand.remove(0);
			int myCardValue = this.gameEvents.getCardValue(myCardEvent);
			if (myCardValue == 0)
				playedZeroCardCount++;
			/*
			 * Publishes it.
			 */
			this.entity.generateEvent(myCardEvent);
			/*
			 * See the cards played by all the players.
			 */
			List<Event> recievedEvents = this.entity.suspend();
			if (recievedEvents.isEmpty())
				throw new IllegalStateException();
			/*
			 * Collects them.
			 */
			for (Event e : recievedEvents)
				playedCards.add(e);
			/*
			 * Am I alone ?
			 */
			if (isFirstCard && recievedEvents.size() == 1) {
				this.trace.println(this.entity.getCommunity().getInstant()
						+ " : " + this.entity + " wins as last player.");
				return false;
			}
			/*
			 * Have I the best card ?
			 */
			int bestCardValue = this.bestCardValue(recievedEvents);
			if (myCardValue != bestCardValue) {
				/*
				 * Loosing instant.
				 */
				if (playedZeroCardCount != 0)
//					this.trace.println(this.reac.getCommunity().getInstant()
//							+ " : " + this.reac + " looses some 'zero' card.")
							;
				this.entity.stop();
				return true;
			}
			int multiplicity = this.cardValueCount(recievedEvents, bestCardValue);
			if (multiplicity == 1) {
				/*
				 * Winning instant.
				 */
				int obtainedZeroCardCount = this.cardValueCount(playedCards, 0);
				if (obtainedZeroCardCount > playedZeroCardCount)
					this.trace.println(this.entity.getCommunity().getInstant()
							+ " : " + this.entity + " obtains some 'zero' card.");
				this.winCards(playedCards);
				this.entity.stop();
				return true;
			}
			/*
			 * Play another card.
			 */
			isFirstCard = false;
		}
	}

	/**
	 * Returns the best card value in a list of card.
	 * @param cardEvents a list of Event.
	 * @return the best card value inside the given list of card.
	 */
	private int bestCardValue(List<Event> cardEvents) {
		int bestCardValue = Integer.MAX_VALUE;
		for (Event e : cardEvents) {
			int card = this.gameEvents.getCardValue(e);
			if (card < bestCardValue)
				bestCardValue = card;
		}
		return bestCardValue;
	}
	
	/**
	 * Returns the count of a given card value in a given list of card.
	 * @param cardEvents a list of Event.
	 * @param cardValue a card value.
	 * @return the count of a given card value in a given list of card.
	 */
	private int cardValueCount(List<Event> cardEvents, int cardValue) {
		int n = 0;
		for (Event e : cardEvents)
			if (this.gameEvents.getCardValue(e) == cardValue)
				n++;
		return n;
	}
	
	///////////////////////
	// Agent implementation
	///////////////////////
	
	/**
	 * Begins to play at the given instant.
	 */
	public void init(LaunchingContext launchingContext) {
		super.init(launchingContext);
		ReactiveCommunity community = launchingContext.getReactiveCommunity();
		while (community.getInstant() < this.startInstant)
			this.entity.stop();
	}

	/**
	 * At each instant, a card (may be more) is played.
	 */
	public void participate() {
		while (this.playACard())
			if (this.entity.getCommunity().getInstant() == 1000) {
				this.trace.println(this.entity.getCommunity().getInstant()
						+ " : " + this.entity + " quits a too long game.");
				return;
			}
	}
}
