/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.market.agent;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;

import org.xmloperator.reactive.demo.market.ProjectedAccount;
import org.xmloperator.reactive.demo.market.Proposal;
import org.xmloperator.reactive.demo.market.Resource;
import org.xmloperator.reactive.demo.market.action.OrderAction;

/**
 * Mission of buying and selling a resource for another for the account of a
 * Transformer.
 */
final class Mission {
	private final long operatorKey;
	final ProjectedAccount purchasedResourceAccount;
	final ProjectedAccount proposedResourceAccount;
	
	private BigInteger resourceUnitCountToPurchase = BigInteger.ZERO;
	
	private final List<OrderAction> orderList = new ArrayList<OrderAction>();
	
	Mission(long operatorKey, ProjectedAccount purchasedResourceAccount,
			ProjectedAccount proposedResourceAccount) {
		if (purchasedResourceAccount == null || proposedResourceAccount == null)
			throw new IllegalArgumentException();
		this.operatorKey = operatorKey;
		this.purchasedResourceAccount = purchasedResourceAccount;
		this.proposedResourceAccount = proposedResourceAccount;
	}

	final boolean concerns(Resource purchasedResource, Resource proposedResource) {
		return this.purchasedResourceAccount.getResource() == purchasedResource
				&& this.proposedResourceAccount.getResource() == proposedResource;
	}
	
	final double virtualRate() {
		return this.proposedResourceAccount.getResource().getUnitValue()
				/ this.purchasedResourceAccount.getResource().getUnitValue();
	}
	
	final void setResourceUnitCountToPurchase(
			BigInteger resourceUnitCountToPurchase) {
		if (resourceUnitCountToPurchase == null)
			throw new IllegalArgumentException();
		this.resourceUnitCountToPurchase = resourceUnitCountToPurchase;
	}
	
	final OrderAction processProposal(Proposal proposal) {
		BigInteger purchasedUnitCount = this.resourceUnitCountToPurchase;
		if (purchasedUnitCount.compareTo(BigInteger.ZERO) == 0)
			return null;
		BigInteger paymentUnitCount =  this.proposedResourceAccount
				.getSoldWithNegativeAnticipations(this.operatorKey);
		BigInteger purchasedFactor = purchasedUnitCount.divide(
				BigInteger.valueOf(proposal
						.getProposedResourceAtomicUnitCount())).add(BigInteger.ONE);
		BigInteger paymentFactor = paymentUnitCount.divide(BigInteger
				.valueOf(proposal.getPaymentResourceAtomicUnitCount()));
		BigInteger factor = purchasedFactor.compareTo(paymentFactor) < 0 ? purchasedFactor
				: paymentFactor;
		if (factor.compareTo(BigInteger.ZERO) <= 0)
			return null;
		if (factor.compareTo(proposal.getAvailableFactor()) > 0)
			factor = proposal.getAvailableFactor();
		OrderAction order = new OrderAction(this.operatorKey, proposal, factor,
				this.proposedResourceAccount, this.purchasedResourceAccount);
		this.orderList.add(order);
		this.proposedResourceAccount.addAnticipation(this.operatorKey,
				order.getNegativeAnticipation());
		this.purchasedResourceAccount.addAnticipation(this.operatorKey,
				order.newPositiveAnticipation());
		return order;
	}
	
	final boolean removeOrder(OrderAction order) {
		if (order == null)
			throw new IllegalArgumentException();
		if (!this.orderList.remove(order))
			return false;
		this.proposedResourceAccount.removeAnticipation(this.operatorKey,
				order.getNegativeAnticipation());
		this.purchasedResourceAccount.removeAnticipation(this.operatorKey, 
				order.newPositiveAnticipation());
		return true;
	}
	
	final boolean isOrderListEmpty() {
		return this.orderList.isEmpty();
	}
	
	////////////////////////
	// Object implementation
	////////////////////////
	
	final boolean equalsTo(Object object) {
		if (!(object instanceof Mission))
			return super.equals(object);
		return ((Mission) object).concerns(this.purchasedResourceAccount
				.getResource(), this.proposedResourceAccount.getResource());
	}
}
