/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.market.agent;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;

import org.xmloperator.reactive.PrintStream;
import org.xmloperator.reactive.demo.market.ProjectedAccount;
import org.xmloperator.reactive.demo.market.MarketBoard;
import org.xmloperator.reactive.demo.market.Proposal;
import org.xmloperator.reactive.demo.market.Resource;
import org.xmloperator.reactive.demo.market.action.OrderAction;
import org.xmloperator.reactive.demo.market.action.DeliveryAction;
import org.xmloperator.reactive.demo.market.impl.ProposalImpl;

/**
 * Resource supply by an Operator.
 */
final class ResourceSupply {
	private static final double TRANSACTION_RELATIVE_PROFIT = 0.02;
	private static final double ACCEPTABLE_RATE_RELATIVE_ERROR = 0.0001;
	private static final int[] PREFERRED_UNIT_COUNTS = new int[] { 1, 2, 3, 4,
			5, 6, 7, 8, 9, 10, 11, 12, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47,
			51, 53, 57, 59, 61, 67, 71, 73, 79, 83, 89, 91, 97 };
	private static final BigInteger[] FRACTION_TO_ENGAGE = new BigInteger[] {
			BigInteger.valueOf(1), BigInteger.valueOf(2) }; // {nominator, denominator}
	
	/**
	 * Search for a fraction (n / d) that approximates a given rate.
	 * @param rate a rate greater than 1.0.
	 * @return a proposed fraction denominator.
	 */
	private static final int bestRateDenominator(double rate) {
		int bestDenominator = 0;
		double bestError = Double.MAX_VALUE;
		for (int denominator : PREFERRED_UNIT_COUNTS) {
			int nominator = (int) Math.round(rate * denominator);
			double approximation = ((double) nominator)
					/ ((double) denominator);
			double error = Math.abs(approximation - rate) / rate;
			if (error < bestError) {
				bestDenominator = denominator;
				bestError = error;
			}
			if (error <= ACCEPTABLE_RATE_RELATIVE_ERROR)
				break;
		}
		return bestDenominator;
	}
	
	final MarketBoard marketBoard;
	final Resource proposedResource;
	final Resource againstResource;
	
	private final long operatorKey;
	private final ProjectedAccount proposedResourceAccount;
	private final ProjectedAccount againstResourceAccount;
	private final List<Campaign> campaignList = new ArrayList<Campaign>();
	private final List<OrderAction> orderList = new ArrayList<OrderAction>();
	private BigInteger accumulationFactor = BigInteger.ZERO;
	private PrintStream trace = null;
	
	ResourceSupply(MarketBoard marketBoard, long operatorKey,
			ProjectedAccount proposedResourceAccount,
			ProjectedAccount againstResourceAccount) {
		if (marketBoard == null || proposedResourceAccount == null
				|| againstResourceAccount == null)
			throw new IllegalArgumentException();
		this.marketBoard = marketBoard;
		this.proposedResource = proposedResourceAccount.getResource();
		this.againstResource = againstResourceAccount.getResource();
		this.operatorKey = operatorKey;
		this.proposedResourceAccount = proposedResourceAccount;
		this.againstResourceAccount = againstResourceAccount;
	}

	void setTrace(PrintStream trace) {
		this.trace = trace;
	}

	/**
	 * Starts a new campaign with a Proposal.
	 * 
	 * @param instant
	 *            the current instant.
	 * @return a Message for announcing the new campaign or null if no new
	 *         campaign.
	 */
	Proposal startNewCampaign(long instant) {
		Proposal proposal = this.buildProposal();
		if (proposal == null)
			return null;
		{
			Campaign campaign = new Campaign(instant, proposal);
			this.campaignList.add(campaign);
		}
		this.proposedResourceAccount.addAnticipation(this.operatorKey,
				proposal.getNegativeAnticipation());
		this.againstResourceAccount.addAnticipation(this.operatorKey,
				proposal.getPositiveAnticipation());
		return proposal;
	}

	/**
	 * An OrderAction is received. It is accepted if and only if it is about
	 * the current Proposal.
	 * 
	 * @param order
	 *            a just received OrderAction.
	 * @param campaignInstant the instant of Campaign starting
	 * @return true if the order is accepted, false elsewhere.
	 */
	boolean addOrder(OrderAction order, long campaignInstant) {
		Campaign campaign = this.getCampaign(campaignInstant);
		if (campaign == null || order.proposal != campaign.proposal)
			return false;
		this.orderList.add(order);
		return true;
	}
	
	/**
	 * Processes the whole orders relative to a Proposal. Returns the corresponding delivery actions.
	 * @param campaignInstant the instant of Campaign starting.
	 * @return the delivery actions corresponding to the waiting orders.
	 */
	DeliveryAction[] processDelivery(long campaignInstant) {
		Campaign campaign = this.getCampaign(campaignInstant);
		if (campaign == null)
			return new DeliveryAction[0];
		BigInteger availableFactor = campaign.proposal.getAvailableFactor();
		int orderCount = this.orderList.size();
		BigInteger claimedFactor = BigInteger.ZERO;
		for (OrderAction order : this.orderList) {
			if (order.proposal != campaign.proposal)
				throw new IllegalStateException();
			claimedFactor = claimedFactor.add(order.orderFactor);
		}
		DeliveryAction[] deliveries = new DeliveryAction[orderCount];
		if (claimedFactor.compareTo(availableFactor) <= 0) {
			/*
			 * Full delivery.
			 */
			this.accumulationFactor = this.accumulationFactor
					.add(claimedFactor);
			for (int k = 0; k < orderCount; k++)
				deliveries[k] = DeliveryAction.fullDelivery(this.operatorKey,
						this.orderList.get(k), this.proposedResourceAccount);
		} else {
			/*
			 * Partial delivery.
			 */
			this.accumulationFactor = this.accumulationFactor
					.add(availableFactor);
			BigInteger[] attributedFactors = new BigInteger[orderCount];
			this.distribute(availableFactor, claimedFactor, attributedFactors);
			for (int k = 0; k < orderCount; k++) {
				OrderAction order = this.orderList.get(k);
				BigInteger factor = attributedFactors[k];
				deliveries[k] = new DeliveryAction(this.operatorKey, order,
						factor, order.orderFactor.subtract(factor),
						this.proposedResourceAccount, order.deliveryAccount,
						this.againstResourceAccount, order.paymentAccount);
			}
		}
		this.orderList.clear();
		this.removeCampaign(campaignInstant);
		return deliveries;
	}
	
	/**
	 * Removes the Campaign starting at the given instant..
	 * @param campaignInstant the instant of Campaign starting.
	 */
	void removeCampaign(long campaignInstant) {
		Campaign campaign = this.getCampaign(campaignInstant);
		if (campaign == null)
			return;
		Proposal proposal = campaign.proposal;
		this.proposedResourceAccount.removeAnticipation(this.operatorKey,
				proposal.getNegativeAnticipation());
		this.againstResourceAccount.removeAnticipation(this.operatorKey,
				proposal.getPositiveAnticipation());
		this.campaignList.remove(campaign);
	}
	
	boolean isCampaignListEmpty() {
		return this.campaignList.isEmpty();
	}
	
	BigInteger getAccumulationFactor() {
		return this.accumulationFactor;
	}
	
	/**
	 * Build a new Proposal.
	 * @return a new proposal or null in case of insufficient stock.
	 */
	private Proposal buildProposal() {
		int nominator;
		int denominator;
		{
			double exchangeRate = (1.0 + TRANSACTION_RELATIVE_PROFIT)
					* this.proposedResource.getUnitValue()
					/ this.againstResource.getUnitValue();
			boolean isInverseRate = exchangeRate < 1.0;
			double rate = isInverseRate ? 1.0 / exchangeRate : exchangeRate;
			denominator = bestRateDenominator(rate);
			nominator = (int) Math.round(rate * denominator);
			if (isInverseRate) {
				int term = nominator;
				nominator = denominator;
				denominator = term;
			}
			if (this.trace != null) {
//				double fraction = ((double) nominator) / ((double) denominator);
//				double refRate = this.proposedResource.getUnitValue()
//						/ this.againstResource.getUnitValue();
//				double relativeProfit = fraction / refRate - 1.0;
//				this.trace.println("BuildProposal(" + this.againstResource
//						+ ">" + this.proposedResource + ") " + nominator + "/"
//						+ denominator + " " + Math.round(1000 * exchangeRate)
//						+ "/1000 profit: " + Math.round(1000 * relativeProfit)
//						+ "/1000");
			}
		}
		BigInteger commitmentToTake = this.proposedResourceAccount
				.getSoldWithNegativeAnticipations(this.operatorKey).multiply(
						FRACTION_TO_ENGAGE[0]).divide(FRACTION_TO_ENGAGE[1]);
		BigInteger availableFactor = commitmentToTake.divide(
				BigInteger.valueOf(denominator));
		return availableFactor.compareTo(BigInteger.ZERO) > 0 ? new ProposalImpl(
				this.marketBoard, this.proposedResource, denominator,
				nominator, availableFactor, this.againstResourceAccount) : null;
	}

	/**
	 * Returns the Campaign that has been started at the given instant.
	 * @param instant the campaign starting instant.
	 * @return the Campaign that has been started at the given instant.
	 */
	private Campaign getCampaign(long instant) {
		for (Campaign campaign : this.campaignList)
			if (campaign.proposalInstant == instant)
				return campaign;
		return null;
	}
	
	/**
	 * Realizes a distribution of the available quantities, which are less than
	 * claimed quantities.
	 * 
	 * @param availableFactor
	 *            the proposed factor, which multiplies the Proposal.
	 * @param claimedFactor
	 *            the sum of claimed factor.
	 * @param attributedFactors
	 *            for output of factors attributed to commands.
	 */
	private void distribute(BigInteger availableFactor,
			BigInteger claimedFactor, BigInteger[] attributedFactors) {
		/*
		 * Main distribution.
		 */
		int orderCount = this.orderList.size();
		for (int k = 0; k < orderCount; k++)
			attributedFactors[k] = this.orderList.get(k).orderFactor
					.multiply(availableFactor).divide(claimedFactor);
		/*
		 * Final distribution at the peak remains.
		 */
		BigInteger obtainedFactor = BigInteger.ZERO;
		for (int k = 0; k < orderCount; k++)
			obtainedFactor = obtainedFactor.add(attributedFactors[k]);
		BigInteger[] remains = new BigInteger[orderCount];
		for (int k = 0; k < orderCount; k++)
			remains[k] = this.orderList.get(k).orderFactor
					.subtract(attributedFactors[k]);
		while (obtainedFactor.compareTo(availableFactor) < 0) {
			int peakIndex = -1;
			BigInteger peakRemains = BigInteger.ZERO;
			for (int k = 0; k < orderCount; k++)
				if (remains[k].compareTo(peakRemains) > 0) {
					peakIndex = k;
					peakRemains = remains[k];
				}
			if (peakIndex < 0)
				throw new IllegalStateException();
			attributedFactors[peakIndex] = attributedFactors[peakIndex]
					.add(BigInteger.ONE);
			remains[peakIndex] = BigInteger.ZERO;
			obtainedFactor.add(BigInteger.ONE);
		}
	}
}
