/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.market.impl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Random;

import org.xmloperator.reactive.demo.market.Ingredient;
import org.xmloperator.reactive.demo.market.Resource;
import org.xmloperator.reactive.demo.market.Transformations;
import org.xmloperator.reactive.demo.market.Transformation;
import org.xmloperator.reactive.demo.market.util.Matrix;

/**
 * Transformations implementation.
 */
public final class TransformationsImpl implements Transformations {
	private final List<TransformationImpl> transformationList = new ArrayList<TransformationImpl>();
	private final List<TransformationImpl> unmodifiableTransformationList = Collections
			.unmodifiableList(this.transformationList);
	private final List<Resource> resourceList = new ArrayList<Resource>();
	private final List<Resource> unmodifiableResourceList = Collections
			.unmodifiableList(this.resourceList);
	private final double resourceBasketValue;
	
	public TransformationsImpl(TransformationImpl[] transformations) {
		if (transformations == null)
			throw new IllegalArgumentException();
		for (TransformationImpl transformation : transformations) {
			if (transformation == null)
				throw new IllegalArgumentException();
			this.transformationList.add(transformation);
			for (Ingredient ingredient : transformation.getIngredientList()) {
				Resource resource = ingredient.getResource();
				if (!this.resourceList.contains(resource))
					this.resourceList.add(resource);
			}
		}
		this.resourceBasketValue = this.computeResourceBasketValue();
	}

	private double computeResourceBasketValue() {
		double basketValue = 0.0;
		double weightSum = 0.0;
		for (Resource resource : this.resourceList) {
			double weight = resource.getWeight();
			basketValue += weight * resource.getUnitValue();
			weightSum += weight;
		}
		basketValue /= weightSum;
		return basketValue;
	}

	/**
	 * Adjusts randomly the coefficients of the transformations.
	 * @param random a random generator.
	 */
	final void adjustTransformations(Random random) {
		for (TransformationImpl transformation : this.transformationList)
			transformation.adjustProcess(random);
	}

	/**
	 * Reduces the variance of relative (to turnover) profits of the
	 * transformations by adjusting the unit values of the resources.
	 * 
	 * @return true in case of success, false if some resource unit value cannot
	 *         be determined.
	 */
	final boolean reduceRelativeProfitsVariance() {
		/*
		 * Profit of transformations can be expressed as follows :
		 * 		G(p) = A(p,n).X(n)
		 * where :
		 * 		G(p) is the relative profit vector,
		 * 		X(n) is the resource unit value vector
		 * 		A(p,n) is a matrix of coefficients.
		 * 		p is the number of transformations.
		 * 		n is the number of resources.
		 * Target G(p) is p times the relative profit average.
		 * The expression to minimize is :
		 * 		E = (A.X - G)t.(A.X - G)
		 * 		E = Xt.At.A.X - 2.Xt.At.G + Gt.G
		 * where At is the matrix A transposed.
		 * Differential of E is :
		 * 		dE = 2.dXt.At.A.X - 2.dXt.At.G
		 * 		dE = 2.dXt.(At.A.X - At.G)
		 * Differential of E is null for :
		 * 		(At.A).X = At.G
		 * which is a linear system with n equations with n unknowns.
		 */
		int p = this.transformationList.size();
		int n = this.resourceList.size();
		if (p < n)
			return false;
		double[] coefs = new double[p * n];
		Arrays.fill(coefs, 0.0);
		for (int r = 0; r < p; r++) {
			Transformation transformation = this.transformationList.get(r);
			double turnover = TransformationUtils.getTurnover(transformation);
			int base = r * n;
			for (Ingredient ingredient : transformation.getIngredientList()) {
				int index = this.resourceList.indexOf(ingredient.getResource());
				if (index < 0)
					throw new IllegalStateException();
				coefs[base + index] = ((double) ingredient.getResourceUnitCount())
						/ turnover;
			}
		}
		Matrix mat_A = new Matrix(p, n);
		mat_A.set(coefs);
		double averageRelativeProfit = this.relativeProfitAverage();
		double[] targets = new double[p];
		Arrays.fill(targets, averageRelativeProfit);
		Matrix mat_G = new Matrix(p, 1);
		mat_G.set(targets);
		Matrix mat_At = mat_A.transpose();
		Matrix mat_M = Matrix.product(mat_At, mat_A);
		Matrix mat_V = Matrix.product(mat_At, mat_G);
		Matrix mat_S = Matrix.resolve(mat_M, mat_V);
		if (mat_S == null)
			return false;
		for (int r = 0; r < n; r++) {
			Resource resource = this.resourceList.get(r);
			double unitValue = mat_S.get(r, 0);
			((ResourceImpl) resource).setUnitValue(unitValue);
		}
		double normalizationFactor = this.resourceBasketValue
				/ this.computeResourceBasketValue();
		for (int r = 0; r < n; r++) {
			Resource resource = this.resourceList.get(r);
			double unitValue = resource.getUnitValue();
			((ResourceImpl) resource).setUnitValue(unitValue
					* normalizationFactor);
		}
		return true;
	}
	
	///////////////////////////////////
	// ResourceEvaluator implementation
	///////////////////////////////////
	
	public final List<? extends Transformation> getTransformationList() {
		return this.unmodifiableTransformationList;
	}

	public double relativeProfitAverage() {
		if (this.transformationList.size() < 1)
			throw new IllegalStateException();
		double average = 0.0;
		for (Transformation transformation : this.transformationList)
			average += TransformationUtils.getRelativeProfit(transformation);
		average /= this.transformationList.size();
		return average;
	}
	
	public double relativeProfitStandardDeviation() {
		if (this.transformationList.size() < 2)
			throw new IllegalStateException();
		double average = this.relativeProfitAverage();
		double variance = 0.0;
		for (Transformation transformation : this.transformationList) {
			double deviation = TransformationUtils
					.getRelativeProfit(transformation)
					- average;
			variance += deviation * deviation;
		}
		variance /= this.transformationList.size() - 1;
		return Math.sqrt(variance);
	}

	public final List<Resource> getResourceList() {
		return this.unmodifiableResourceList;
	}

	public double getResourceBasketValue() {
		return this.resourceBasketValue;
	}
}
